#------------------------------------------------------------------------------------------------
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#------------------------------------------------------------------------------------------------
#	This plugin was developed and tested in a Windows environment and may not work on a Mac.
#------------------------------------------------------------------------------------------------
#  Name:	PointOfTangency
#    By:	sdmitch
# Usage:	To calculate the points of tangency. 
#  Note:	Select the circle or arc then a point.  Select large circle then small circle.  If Arc
#					or Circle option is chosen, select two circles or arcs to place arc or circle tangent to.
#					This is a 2D app and only works on the XY plane.
#  Date:	Dec 2011
#------------------------------------------------------------------------------------------------

	require 'sketchup'

	if not file_loaded?(File.basename(__FILE__))
		UI.menu("Draw").add_item("Tangent Geometry") { Sketchup.active_model.select_tool PointOfTangency.new }  #MV Changed menu from plugin to draw and "tangent Geometry"
		

	
	#####   Toolbar	
	toolbar = UI::Toolbar.new("Tangent Points")

	cmd = UI::Command.new("Tangent Points") { Sketchup.active_model.select_tool PointOfTangency.new }
	cmd.small_icon = File.join(File.dirname(__FILE__), "tangent_points", "tp16.png")
	cmd.large_icon = File.join(File.dirname(__FILE__), "tangent_points", "tp24.png")
	cmd.status_bar_text = cmd.tooltip = "Draw Tangent Geometry"
	toolbar.add_item(cmd)
	
	toolbar.restore
### End toolbar block
	end
	
file_loaded(File.basename(__FILE__))
###-------------------------------------------------------------------
	
class PointOfTangency

	def initialize
		@ip = Sketchup::InputPoint.new
		@ip1 = Sketchup::InputPoint.new
		@mod = Sketchup.active_model
		@ent = @mod.active_entities
		@sel = @mod.selection
		Sketchup.send_action "viewTop:"
		Sketchup.send_action "viewZoomExtents:"
		@mod.active_view.refresh
		@typ="Line" if !@typ
		@rad=2.feet if !@rad
		self.dialog
		self.reset
	end

	def dialog
		ans=inputbox(["Add Tangent:"],[@typ],["Arc|Circle|Line"],"Tangent to two Arcs/Circles")
		if ans
			@typ=ans[0]
			if @typ != "Line"
				ans=inputbox(["Radius:"],[@rad],"Tangent Arc/Circle radius")
				if ans
					@rad=ans[0].to_l
				else
					onCancel(nil,nil)
				end
			end
		else
			onCancel(nil,nil)
		end
	end
	
	def reset
		@c2 = nil
		@state = 0
		@best = nil
		@ip1.clear
		@sel.clear
		@drawn = false
	end

	def set_current_point(x, y, view)
		if( !@ip.pick(view, x, y, @ip1) )
			return false
		end
		@ip1.copy! @ip
		@ip0 = @ip.position
		view.lock_inference
		view.tooltip = @ip.tooltip
		view.tooltip = "Pick Center Point" if @state == 2
		if @typ == "Line"
			Sketchup::set_status_text("Select big circle or right click to exit", SB_PROMPT) if @state == 0
			Sketchup::set_status_text("Select sm circle/point or right click to start over", SB_PROMPT) if @state == 1
			Sketchup::set_status_text("", SB_VCB_LABEL)
			Sketchup::set_status_text("", SB_VCB_VALUE)
		else
			Sketchup.set_status_text("Select first arc/circle.  Press Shift to toggle Arc/Circle.  Press Esc key to exit.",SB_PROMPT) if @state == 0
			Sketchup.set_status_text("Select second arc/circle.  Press Shift to toggle Arc/Circle.  Press Esc key to exit.",SB_PROMPT) if @state == 1
			Sketchup.set_status_text("Select center point to use.  Press Shift to toggle Arc/Circle.  Press Esc key to exit.",SB_PROMPT) if @state == 2
			Sketchup.set_status_text("Tangent #{@typ} Radius:",SB_VCB_LABEL)
			Sketchup.set_status_text(Sketchup.format_length(@rad),SB_VCB_VALUE)
		end
	end

	def onMouseMove(flags, x, y, view)
		self.set_current_point(x, y, view)
		ph = view.pick_helper;ph.do_pick(x, y)
		@best = ph.best_picked if ph.best_picked != @best
		view.invalidate
	end

	def onLButtonDown(flags, x, y, view)
		self.set_current_point(x, y, view)
		case @state
			when 0
				if @best.class==Sketchup::Edge && @best.curve && @best.curve.class==Sketchup::ArcCurve
					@c1 = @best.curve; @sel.add @c1.edges; @state += 1
					@r1=@c1.radius;@cp1=@c1.center
				end
			when 1
				@mod.start_operation "Tangents"
				if @best.class==Sketchup::Edge && @best.curve && @best.curve.class==Sketchup::ArcCurve
					@c2 = @best.curve; @sel.add @c2.edges; @state += 1
					@r2=@c2.radius;@cp2=@c2.center
					if @typ == "Line"
						if @r1>@r2 then
							r3=@r1-@r2
							edges=@ent.add_circle(@cp1,Z_AXIS,r3)
							c3=edges[0].curve
							pt1,pt2=self.pot(c3,@cp2)
							if pt1 then
								vec=@cp1.vector_to(pt1)
								p1=@cp1.offset(vec,@r1)
								p2=@cp2.offset(vec,@r2)
								@ent.add_cpoint(p1); 
								@ent.add_cpoint(p2); 
								@ent.add_cline(p1,p2)
								vec=@cp1.vector_to(pt2)
								p1=@cp1.offset(vec,@r1)
								p2=@cp2.offset(vec,@r2)
								@ent.add_cpoint(p1); 
								@ent.add_cpoint(p2); 
								@ent.add_cline(p1,p2)
							end
							@ent.erase_entities(edges)
						end
					else
						@cp3,@cp4=self.circle_circle_intersection(@cp1,@r1+@rad,@cp2,@r2+@rad)
						if @cp3
							@ent.add_cpoint(@cp3);@ent.add_cpoint(@cp4)
							@state = 2
						else
							self.reset
						end
					end
				else
					cpt = @ip.position
					pt1,pt2=self.pot(@c1,cpt)
					if pt1 then
						@ent.add_cpoint(pt1); @ent.add_cline(cpt,pt1)
						@ent.add_cpoint(pt2); @ent.add_cline(cpt,pt2)
					end
				end
				(@mod.commit_operation;self.reset) if @state == 1
			when 2
				keep = @ip.position
				if keep.distance(@cp3) < keep.distance(@cp4)
					@cp4=nil;self.add_arc_or_circle(@cp3,@cp4)
				else
					@cp3=nil;self.add_arc_or_circle(@cp3,@cp4)
				end
				@mod.commit_operation;self.reset
		end
		view.lock_inference
	end
	
	def pot(cir,p1)
		p0 = cir.center; rad = cir.radius
		hyp = p0.distance(p1)
		if rad < hyp
			bas = Math.sqrt(hyp**2 - rad**2)
			vec=p1.vector_to(p0); az=self.vector_to_azimuth(vec)
			sin=rad/hyp; ang=Math.asin(sin)
			azm = az - ang; vec=self.azimuth_to_vector(azm)
			pt1=p1.offset(vec,bas);
			azm = az + ang; vec=self.azimuth_to_vector(azm)
			pt2=p1.offset(vec,bas); 
			return pt1,pt2
		else
			UI.messagebox "Point must be outside of the circle"
			return nil,nil
		end
	end
	
	def  vector_to_azimuth(v)
	# --------------------------------------------------------------------------------
	#	Compute Azimuth based on positive x-axis as 0 degrees
	# --------------------------------------------------------------------------------
		v.normalize!
		va=Math::acos(v.z) # Vertical angle
		if v.x < 0
			az = Math::PI - Math::asin(v.y) # Quadrant III or IV
		elsif v.y < 0
			az = Math::PI * 2 + Math::asin(v.y) # Quadrant II
		else
			az = Math::asin(v.y) # Quadrant I
		end
		return az # Horizontal angle
	end
	
	def azimuth_to_vector(a)
		v=Geom::Vector3d.new(Math.cos(a),Math.sin(a),0)
	end

	def add_arc_or_circle(cp3,cp4)

		if cp3
			vec1=cp3.vector_to @cp1;vec2=cp3.vector_to @cp2
			tp1=cp3.offset(vec1,@rad);tp2=cp3.offset(vec2,@rad)
			@ent.add_cpoint(tp1);@ent.add_cpoint(tp2)
			if @typ == "Arc"
				vec1=cp3.vector_to(@cp1);vec2=cp3.vector_to(@cp2);ang=vec1.angle_between(vec2)
				@ent.add_arc(cp3,vec1,@c1.normal,@rad,0,ang,15)
			else
				edges=@ent.add_circle(cp3,@c1.normal,@rad,30)
				face=@ent.add_face(edges)
			end
		end
		if cp4
			vec1=cp4.vector_to @cp2;vec2=cp4.vector_to @cp1
			tp1=cp4.offset(vec1,@rad);tp2=cp4.offset(vec2,@rad)
			@ent.add_cpoint(tp1);@ent.add_cpoint(tp2)
			if @typ == "Arc"
				vec1=cp4.vector_to(@cp2);vec2=cp4.vector_to(@cp1);ang=vec1.angle_between(vec2)
				@ent.add_arc(cp4,vec1,@c2.normal,@rad,0,ang,15)
			else
				edges=@ent.add_circle(cp4,@c2.normal,@rad,30)
				face=@ent.add_face(edges)
			end
		end

	end
	
	def circle_circle_intersection(c1,r1,c2,r2)
		dx=c2.x-c1.x;dy=c2.y-c1.y;dz=c2.z-c1.z
		d=Math.sqrt(dx**2 + dy**2)
		if d < (r1 + r2) and d != 0.0 and dz.abs < 0.001
			# Determine the distance to point 3, the intersection of lines between centers and intersection points.
			a = ((r1*r1) - (r2*r2) + (d**2)) / (2.0 * d)
			# Determine the coordinates of point 3.
			x2 = c1.x + (dx * a/d);y2 = c1.y + (dy * a/d);
			# Determine the distance from point 3 to either of the intersection points.
			h = Math.sqrt((r1*r1) - (a*a));
			# Now determine the offsets of the intersection points from point 3.
			rx = -dy * (h/d);ry = dx * (h/d);
			# Determine the absolute intersection points.
			x = x2 + rx;y = y2 + ry;ip1=Geom::Point3d.new(x,y,0)
			x = x2 - rx;y = y2 - ry;ip2=Geom::Point3d.new(x,y,0)
			return ip1,ip2
		else
			UI.messagebox "Invalid input for circle_circle intersect, d=#{d}, dz=#{dz}"
			return nil,nil
		end
	end

	def onKeyDown(key, repeat, flags, view)
		puts "onKeyDown: key = " + key.to_s 
		case key
			when VK_SHIFT
				if @typ == "Arc"
					@typ = "Circle"
				else
					@typ = "Arc"
				end
				Sketchup.set_status_text("Tangent #{@typ} Radius:",SB_VCB_LABEL)
				Sketchup.set_status_text(Sketchup.format_length(@rad),SB_VCB_VALUE)
		end
	end
	
	def onUserText(text, view)
		begin
			val = text.to_l
		rescue
			puts "Text = #{text}, view = #{view}"
			return
		end
		@rad = val
		Sketchup.set_status_text("Tangent #{@typ} Radius:",SB_VCB_LABEL)
		Sketchup.set_status_text(Sketchup.format_length(@rad),SB_VCB_VALUE)
	end
	
	def onCancel(flag, view)
		view.invalidate if @drawn
		Sketchup.send_action "selectSelectionTool:"
	end
	
	def onRButtonDown(flags,x,y,view)
		self.dialog
		self.reset
	end

	def draw(view)
		@drawn = false
		if( @ip.valid? && @ip.display? )
			@ip.draw(view)
			@drawn = true
		end
		
		if @best.class==Sketchup::Edge && @best.curve && @best.curve.class==Sketchup::ArcCurve
			view.line_width=3
			view.drawing_color="orangered"
			verts=@best.curve.vertices
			verts.collect!{|v| v.position}
			if @best.curve.end_angle < 360.degrees
				view.draw_polyline verts
			else
				view.draw GL_LINE_LOOP, verts
			end
		end
	end

end # of class PointOfTangency

